#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef _WIN32
#include <windows.h>
#endif
#include "cdirip.h"

// Linux add-ons by Pedro Melo

#ifndef _WIN32
#include <limits.h>

#define MAX_PATH PATH_MAX
#endif

#ifdef __BORLANDC__
#define strcasecmp stricmp
#endif


int main( int argc, char **argv )
{

const char TRACK_START_MARK[10] = { 0, 0, 0x01, 0, 0, 0, 0xFF, 0xFF, 0xFF, 0xFF };
char       start_mark[10];

int  i;
char cuesheetname[13], filename[MAX_PATH];
FILE *fsource, *fcuesheet;

image_s image = { 0, };
track_s track = { 0, };
opts_s  opts  = { 0, };

image.global_current_session = 0;
track.global_current_track   = 0;
track.position               = 0;

// Opciones

printf("CDIrip 0.5a - (C) 2001 by DeXT\n\n");

opts.ask_for_image = 1;

#ifndef _WIN32
opts.convert = 1;  // Linux only!
#endif

if (argc >= 2)
   {
   for(i = 1; i < argc; i++)
     {
     if (!strncmp(argv[i], "-", 1))
      {
      if (!strcasecmp(argv[i],"-iso"   )) opts.convert  = 1;
      if (!strcasecmp(argv[i],"-bin"   )) opts.convert  = 2;
      if (!strcasecmp(argv[i],"-raw"   )) opts.rawaudio = 1;
      if (!strcasecmp(argv[i],"-info"  )) opts.showinfo = 1;
      if (!strcasecmp(argv[i],"-cut"   )) opts.cutfirst = 1;
      if (!strcasecmp(argv[i],"-cutall")) opts.cutall   = 1;
      if (!strcasecmp(argv[i],"-full"  )) opts.fulldata = 1;
      if (!strcasecmp(argv[i],"-swap"  )) opts.swap     = 1;

      if (!strcasecmp(argv[i],"-cdrecord"  )) { opts.cutall = 1; opts.convert = 1; }
      if (!strcasecmp(argv[i],"-winoncd"   )) { opts.cutall = 1; opts.convert = 1; opts.rawaudio = 1; }
      if (!strcasecmp(argv[i],"-fireburner")) { opts.cutall = 1; opts.convert = 2; }
      }
     else if (i == 1) opts.ask_for_image = 0;
     }
   }

if (opts.ask_for_image == 1)
   {
   if (askfilename(filename) != 0)
      {
      printf("Exit requested by user");
#ifdef _WIN32
      FreeConsole();
      ExitProcess(0);
#else
      exit (EXIT_SUCCESS);
#endif
      }
   }
else
   strcpy(filename, argv[1]);

// Abrir fichero

printf("Searching file: '%s'\n",filename);

fsource = fopen(filename,"rb");

if (fsource == NULL)
   {
   if (errno == ENOENT)  // "No such file or directory"
      {
      strcat(filename, ".cdi");
      printf("Not found. Trying '%s' instead...\n",filename);
      fsource = fopen(filename,"rb");
      }
   }

if (fsource == NULL) error_exit(ERR_OPENIMAGE, filename);

errno = 0;
printf("Found image file. Opening...\n");

fseek(fsource, 0L, SEEK_END);
image.length = ftell(fsource);
fseek(fsource, image.length-8, SEEK_SET);
fread(&image.version, 4, 1, fsource);
fread(&image.header_offset, 4, 1, fsource);

if (image.length < 8) error_exit(ERR_GENERIC, "Image file is too short");

if (errno != 0) error_exit(ERR_READIMAGE, filename);

if (image.header_offset == 0) error_exit(ERR_GENERIC, "Bad image format");

if (image.version == 0x80000004)
   printf("This is a v2 image\n");
else
   if (image.version == 0x80000005)
      printf("This is a v3 image\n");
   else
      error_exit(ERR_GENERIC, "Unsupported image version");

//printf("Header offset: %ld\n",image.header_offset);

fseek(fsource, image.header_offset, SEEK_SET);

fread(&image.sessions, 2, 1, fsource);

printf("\nAnalyzing image...\n");

if (image.sessions == 0)  error_exit(ERR_GENERIC, "Bad format: Could not find header");

printf("Found %d session(s)\n",image.sessions);

image.remaining_sessions = image.sessions;

/////////////////////////////////////////////////////////////// Bucle sessions

while(image.remaining_sessions > 0)
  {
  image.global_current_session++;

  fread(&image.tracks, 2, 1, fsource);
  fseek(fsource, 4, SEEK_CUR);
  image.header_position = ftell(fsource);

  printf("\nSession %d has %d track(s)\n",image.global_current_session,image.tracks);

  if (image.tracks == 0)
     printf("Open session\n");
  else
    {

// Decidir si crear cuesheet

    if (opts.showinfo == 0)
       {
       if (image.global_current_session == 1)
          {
          if (ask_type(fsource, image.header_position) == 2)
             {
             if (opts.convert == 1)
                opts.create_cuesheet = 0;  // Si "/iso" y Modo2 -> no cuesheet
             else
                {
                opts.create_cuesheet = 1;
                opts.convert = 2;          // Si Modo2 -> cuesheet + BIN
                }
             }
          else
             opts.create_cuesheet = 1;     // Si Audio o Modo1
          }
       else
          {
          if (opts.convert == 2)           // ojo! tb depende de lo anterior
             opts.create_cuesheet = 1;
          else
             opts.create_cuesheet = 0;
          }
       }
    else
       opts.create_cuesheet = 0;

// Crear cuesheet

    if (opts.create_cuesheet == 1)
       {
       printf("Creating cuesheet...\n");
       if (image.global_current_session == 1)
          sprintf(cuesheetname,"tdisc.cue");
       else
          sprintf(cuesheetname,"tdisc%d.cue",image.global_current_session);
       fcuesheet = fopen(cuesheetname,"wb");
       }

    image.remaining_tracks = image.tracks;

///////////////////////////////////////////////////////////////// Bucle tracks

    while(image.remaining_tracks > 0)
         {
         track.global_current_track++;
         track.number = image.tracks - image.remaining_tracks + 1;

// Leer datos

         fread(&start_mark, 10, 1, fsource);
         if (memcmp(TRACK_START_MARK, start_mark, 10)) error_exit(ERR_GENERIC, "Unsupported format: Could not find the track start mark");

         fread(&start_mark, 10, 1, fsource);
         if (memcmp(TRACK_START_MARK, start_mark, 10)) error_exit(ERR_GENERIC, "Unsupported format: Could not find the track start mark");

         fseek(fsource, 4, SEEK_CUR);
         fread(&track.filename_length, 1, 1, fsource);
         fseek(fsource, track.filename_length, SEEK_CUR);
         fseek(fsource, 25, SEEK_CUR);
         fread(&track.pregap_length, 4, 1, fsource);
         fread(&track.length, 4, 1, fsource);
         fseek(fsource, 6, SEEK_CUR);
         fread(&track.mode, 4, 1, fsource);
         fseek(fsource, 12, SEEK_CUR);
         fread(&track.start_lba, 4, 1, fsource);
         fread(&track.total_length, 4, 1, fsource);
         fseek(fsource, 16, SEEK_CUR);
         fread(&track.sector_size_value, 4, 1, fsource);

         switch(track.sector_size_value)
               {
               case 0 : track.sector_size = 2048; break;
               case 1 : track.sector_size = 2336; break;
               case 2 : track.sector_size = 2352; break;
               default: error_exit(ERR_GENERIC, "Unsupported sector size");
               }

         if (track.mode > 2) error_exit(ERR_GENERIC, "Unsupported format: Track mode not supported");

         fseek(fsource, 33, SEEK_CUR);
         if (image.version == 0x80000005) fseek(fsource, 9, SEEK_CUR);

         image.header_position = ftell(fsource);

// Mostrar info

         if (opts.showinfo == 0) printf("Saving  ");
         printf("Track: %2d  ",track.global_current_track);
         printf("Type: ");
         switch(track.mode)
               {
               case 0 : printf("Audio/"); break;
               case 1 : printf("Mode1/"); break;
               case 2 :
               default: printf("Mode2/");
               }
         printf("%d  ",track.sector_size);
//       printf("Pregap: %ld  ",track.pregap_length);
         printf("Size: %-6ld  ",track.length);
         printf("LBA: %-6ld",track.start_lba);

         if (opts.showinfo != 0) printf("\n");

//       if (track.pregap_length != 150) printf("Warning! This track seems to have a non-standard pregap...\n");

// Decidir si cortar

         if (opts.fulldata == 0 && track.mode != 0 && image.remaining_sessions > 1)
            opts.do_cut = 2;          // Cortar si pista de datos en 1 sesion
         else
            opts.do_cut = (opts.cutall & !(track.mode != 0 && opts.fulldata == 1))*2 +
                          (opts.cutfirst & track.global_current_track == 1)*2;

// Decidir si convertir

         if (track.mode != 0 && track.sector_size != 2048)
            switch (opts.convert)
                   {
                   case 2 : opts.do_convert = 0; break;
                   case 1 : opts.do_convert = 1; break;
                   case 0 :
                   default: if (track.mode == 1)
                               opts.do_convert = 1;          // Modo1/2352 -> ISO
                            else
                               if (image.global_current_session > 1)
                                  opts.do_convert = 1;       // Modo2 2 sesion -> ISO
                               else
                                  opts.do_convert = 0;       // Modo2 1 sesion -> BIN (obsoleto)
                   }
         else
            opts.do_convert = 0;

         if (track.sector_size == 2048 || (track.mode != 0 && opts.do_convert == 1))
            opts.save_as_iso = 1;
         else
            opts.save_as_iso = 0;

// Guardar la pista

         if (opts.showinfo == 0)
            {
            if (track.total_length < track.length + track.pregap_length)
               {
               printf("\nThis track seems truncated. Skipping...");
               fseek(fsource, track.position, SEEK_SET);
               fseek(fsource, track.total_length, SEEK_CUR);
               track.position = ftell(fsource);
               }
            else
               {
               savetrack(fsource, &track, &opts);
               track.position = ftell(fsource);

// Generar entradas de cuesheet

               if (opts.create_cuesheet == 1 &&
                  !(track.mode == 2 && opts.do_convert == 1))  // No generar entrada si convertimos (obsoleto)
                  savecuesheet(fcuesheet, &track, &opts);

               }
            }

         fseek(fsource, image.header_position, SEEK_SET);


// Cerrar bucles

         image.remaining_tracks--;
         }

    if (opts.create_cuesheet == 1) fclose(fcuesheet);
    }

  fseek(fsource, 8, SEEK_CUR);
  if (image.version == 0x80000005) fseek(fsource, 1, SEEK_CUR);

  image.remaining_sessions--;
  }

// Finalizar

printf("\nAll done!\n");

if (opts.showinfo == 0) printf("Good burnin'...\n");

fclose(fsource);

return 0;
}


//////////////////////////////////////////////////////////////////////////////


void savetrack(FILE *fsource, track_s *track, opts_s *opts)
{

unsigned long i, ii;
unsigned long image_length, track_length;
char cvalue;
char buffer[2352], filename [13];
FILE *fdest;

     fseek(fsource, 0L, SEEK_END);   // chapuza
     image_length = ftell(fsource);  //

     fseek(fsource, track->position, SEEK_SET);

     if (track->mode == 0)
        switch(opts->rawaudio)
              {
              case 1 : sprintf(filename,"taudio%02d.raw",track->global_current_track); break;
              case 0 :
              default: sprintf(filename,"taudio%02d.wav",track->global_current_track);
              }
     else
        switch(opts->save_as_iso)
              {
              case 1 : sprintf(filename,"tdata%02d.iso",track->global_current_track); break;
              case 0 :
              default: sprintf(filename,"tdata%02d.bin",track->global_current_track);
              }

     fdest = fopen(filename,"wb");

     if (fdest == NULL) error_exit(ERR_SAVETRACK, filename);

     setvbuf(fdest, NULL, _IOFBF, 65536);

     fseek(fsource, track->pregap_length*track->sector_size, SEEK_CUR);

     if (opts->do_cut != 0) printf("  [cut: %d]", opts->do_cut);
     track_length = track->length - opts->do_cut;                // para no modificar valor original

     if (track->mode != 0 && track->sector_size != 2048 && opts->do_convert == 1) // convertir
        {
        printf("  [ISO]\n");
        if (track->sector_size == 2336) // Modo2/2336 -> 2048
           {
           for(i = 0; i < track_length; i++)
              {
              if (!(i%128)) show_counter (i, track_length, image_length, ftell(fsource));
              fseek(fsource, 8, SEEK_CUR);
              fread(buffer, 2048, 1, fsource);
              if (!(fwrite(buffer, 2048, 1, fdest))) error_exit(ERR_SAVETRACK, filename);
              fseek(fsource, 280, SEEK_CUR);
              }
           }
        else                            // Modo2/2352 -> 2048
           {
           for(i = 0; i < track_length; i++)
              {
              if (!(i%128)) show_counter (i, track_length, image_length, ftell(fsource));
              fseek(fsource, 16, SEEK_CUR);
              fseek(fsource, 8, SEEK_CUR);
              fread(buffer, 2048, 1, fsource);
              if (!(fwrite(buffer, 2048, 1, fdest))) error_exit(ERR_SAVETRACK, filename);
              fseek(fsource, 280, SEEK_CUR);
              }
           }
        }
     else    // no convertir
        {
        printf("\n");
        if (track->mode == 0)           // Audio
           {
           if (opts->rawaudio == 0) writewavheader(fdest, track_length);
           for(i = 0; i < track_length; i++)
              {
              if (!(i%128)) show_counter (i, track_length, image_length, ftell(fsource));
              fread(buffer, track->sector_size, 1, fsource);
              if (opts->swap)
                 //for (ii = track->sector_size; ii > 0; --ii,--ii)
                 for (ii = 0; ii < track->sector_size; ++ii, ++ii)
                     {
                     cvalue = buffer[ii];
                     buffer[ii] = buffer[ii+1];
                     buffer[ii+1] = cvalue;
                     }
              if (!(fwrite(buffer, track->sector_size, 1, fdest))) error_exit(ERR_SAVETRACK, filename);
              }
           }
        else                            // Datos (raw/no conversion)
           {
           for(i = 0; i < track_length; i++)
              {
              if (!(i%128)) show_counter (i, track_length, image_length, ftell(fsource));
              fread(buffer, track->sector_size, 1, fsource);
              if (!(fwrite(buffer, track->sector_size, 1, fdest))) error_exit(ERR_SAVETRACK, filename);
              }
           }
        }

     if (opts->do_cut != 0) fseek(fsource, opts->do_cut*track->sector_size, SEEK_CUR);

     printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
            "                            "
            "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");

     fflush(fdest);
     fclose(fdest);
}

void show_counter (unsigned long i, unsigned long track_length, unsigned long image_length, long pos)
{
	unsigned long progress, total_progress;

	progress = (i*100/track_length);
	total_progress = ((pos>>10)*100)/(image_length>>10);
	printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
	"[Current: %02d%%  Total: %02d%%]  ",progress, total_progress);
}

//////////////////////////////////////////////////////////////////////////////


void savecuesheet(FILE *fcuesheet, track_s *track, opts_s *opts)
{

char track_format_string[10];

     if (opts->swap == 1)
        strcpy(track_format_string,"MOTOROLA");
     else
        strcpy(track_format_string,"BINARY");

     if (track->mode == 0)
        {
        switch(opts->rawaudio)
              {
              case 1 : fprintf(fcuesheet, "FILE TAUDIO%02d.RAW %s\r\n"
                                          "  TRACK %02d AUDIO\r\n",
                                          track->global_current_track,
                                          track_format_string,
                                          track->number);
                                          break;
              case 0 :
              default: fprintf(fcuesheet, "FILE TAUDIO%02d.WAV WAVE\r\n"
                                          "  TRACK %02d AUDIO\r\n",
                                          track->global_current_track,
                                          track->number);
              }

        if (track->global_current_track > 1)
           fprintf(fcuesheet, "    PREGAP 00:02:00\r\n");
        }
     else
        {
        switch(opts->save_as_iso)
              {
              case 1 : fprintf(fcuesheet, "FILE TDATA%02d.ISO BINARY\r\n"
                                          "  TRACK %02d MODE%d/2048\r\n",
                                          track->global_current_track,
                                          track->number,
                                          track->mode);
                                          break;
              case 0 :
              default: fprintf(fcuesheet, "FILE TDATA%02d.BIN BINARY\r\n"
                                          "  TRACK %02d MODE%d/%d\r\n",
                                          track->global_current_track,
                                          track->number,
                                          track->mode,
                                          track->sector_size);
              }
        }

     fprintf(fcuesheet, "    INDEX 01 00:00:00\r\n");
}


//////////////////////////////////////////////////////////////////////////////


void writewavheader(FILE *fdest, unsigned long track_length)
{

unsigned long      wHeader_length;
unsigned long      wData_length;
unsigned long      wHeaderLength = 16;
unsigned short int wFormat = 1;
unsigned short int wChannels = 2;
unsigned long      wSampleRate = 44100;
unsigned long      wBitRate = 176400;
unsigned short int wBlockAlign = 4;
unsigned short int wBitsPerSample = 16;

      wData_length = track_length*2352;
      wHeader_length = wData_length + 8 + 16 + 12;

      fwrite("RIFF", 4, 1, fdest);
      fwrite(&wHeader_length, 4, 1, fdest);
      fwrite("WAVE", 4, 1, fdest);
      fwrite("fmt ", 4, 1, fdest);
      fwrite(&wHeaderLength, 4, 1, fdest);
      fwrite(&wFormat, 2, 1, fdest);
      fwrite(&wChannels, 2, 1, fdest);
      fwrite(&wSampleRate, 4, 1, fdest);
      fwrite(&wBitRate, 4, 1, fdest);
      fwrite(&wBlockAlign, 2, 1, fdest);
      fwrite(&wBitsPerSample, 2, 1, fdest);
      fwrite("data", 4, 1, fdest);
      fwrite(&wData_length, 4, 1, fdest);
}


//////////////////////////////////////////////////////////////////////////////


unsigned long ask_type(FILE *fsource, long header_position)
{

unsigned char filename_length;
unsigned long track_mode;

    fseek(fsource, header_position, SEEK_SET);
    fseek(fsource, 24, SEEK_CUR);
    fread(&filename_length, 1, 1, fsource);
    fseek(fsource, filename_length, SEEK_CUR);
    fseek(fsource, 39, SEEK_CUR);
    fread(&track_mode, 4, 1, fsource);
    fseek(fsource, header_position, SEEK_SET);
    return (track_mode);
}


//////////////////////////////////////////////////////////////////////////////


void error_exit(long errcode, char *string)
{

char errmessage[256];

    if (errcode != ERR_GENERIC)
       printf("\n%s: %s\n", string, strerror(errno));  // string is used as Filename

    switch(errcode)
          {
          case ERR_OPENIMAGE:
                   strcpy(errmessage, "File not found\n\nYou may have typed a wrong name or path to source CDI image");
                   break;
          case ERR_SAVETRACK:
                   strcpy(errmessage, "Could not save track");
                   break;
          case ERR_READIMAGE:
                   strcpy(errmessage, "Error reading image");
                   break;
          case ERR_GENERIC:
          default: strcpy(errmessage, string);          // string is used as Error message
          }

#ifdef _WIN32
    MessageBoxA(NULL, errmessage, NULL, MB_OK | MB_ICONERROR);
    ExitProcess(0);
#else
    printf(errmessage);
    exit(EXIT_FAILURE);
#endif

}


//////////////////////////////////////////////////////////////////////////////


int askfilename(char *string)
{

#ifdef _WIN32

     printf("\nPlease enter name of CDI image (return to exit): ");

     gets(string);
     
     if (strlen(string) == 0) return 1;

#else
     printf("\nPlease enter name of CDI image (return to exit): ");

     fgets(string, MAX_PATH, stdin);   // Linux doesn't seem lo like much gets()

     if (strlen(string) == 0) return 1;
#endif

     return 0;
}
